
// cbmExport class
var cbmExport = (function () {

	// Private
	var textFile = null;
	var charIndex = 0;
	var tileIndex = 0;
	var tileLayoutIndex = 0;
	var screenIndex = 0;
	var colourIndex = 0;
	var mode = "";
	var inputScreen = 0;
	var _rle = false;
	var _exportBuffer;
	var charImportBuffer = [];
	var singleScreenOnly = false;
	var inputRadix;
	var itemsPerLine = 8;
	var importTileRow = 0;

	// Private methods
	var Load = function (text) {};

	// Enable starts with
	if (typeof String.prototype.startsWith != 'function') {
		// see below for better implementation!
		String.prototype.startsWith = function (str) {
			return this.indexOf(str) == 0;
		};
	}

	var RLEEncode8 = function (sourceArray) {

		var outputList = [];
		var count = 1;
		var current = sourceArray[0];

		var rleOutput = function () {
			var addition = 0;

			// If we have more than one item or the current item cannot be represented as a single value (> 127)
			if (count > 1 || current >= 128) {
				outputList.push(count);
			} else {
				addition = 128;
			}
			outputList.push(current + addition);
		};

		var dumpOutput = function () {
			rleOutput();
			count = 1;
			current = loopItem;
		};

		for (var i = 1; i < sourceArray.length; i++) {
			var loopItem = sourceArray[i];

			if (count > 126) // Max size exceeded
			{
				dumpOutput();
			} else if (current == loopItem) // Tally
			{
				count++;
			} else // New char force output
			{
				dumpOutput();
			}
		}

		dumpOutput();
		return outputList;
	};

	function ExtractItemValue(value) {
		var parseText = value;
		var converted = value;

		try {

			if (value.toString().startsWith('$')) {
				inputRadix = 2;
				parseText = "0x" + value.substring(1);
				converted = parseInt(parseText, 16);
			} else if (value.toString().startsWith('0x')) {
				inputRadix = 1;
				converted = parseInt(parseText, 16);
			} else if (value.toString().startsWith('%')) {
				inputRadix = 3;
				parseText = value.substring(1);
				converted = parseInt(parseText, 2);
			} else {
				inputRadix = 0;
				converted = parseInt(value);
			}
		} catch (err) {
			throw (err);
		}

		return converted;

	}

	function ProcessLineItem(index, value, lineArray) {
		var curScreen;

		if (!singleScreenOnly) {
			curScreen = CbmScreens[inputScreen];
		} else {
			curScreen = CbmScreen;
		}

		//console.log("Processing data cell :" +  index + "value:" + value);
		var itemValue = ExtractItemValue(value);

		switch (mode) {
		case "CHAR":
			charImportBuffer[charIndex] = itemValue;
			charIndex++;
			//console.log("Adding char:" + charIndex + "value: " +  itemValue);
			break;
		case "SCREEN":
			if (singleScreenOnly && inputScreen > 0)
				return;
			curScreen.setCellCharFromIndex(screenIndex, itemValue);
			screenIndex++;
			//console.log("Adding cell:" + screenIndex + "value: " +  itemValue);
			break;
		case "COLOUR":
			if (singleScreenOnly && inputScreen > 0)
				return;
			curScreen.setCellColourFromIndex(colourIndex, itemValue);
			colourIndex++;
			//console.log("Adding colour cell:" + colourIndex + "value: " +  itemValue);
			break;
		case "TILES":
			if (!lineArray.tileProcessed){
				var tile = _CbmTileSet.newTile();
				var arr = lineArray.slice(0).map(ExtractItemValue);
				tile.setFromArray(arr);
				lineArray.tileProcessed = true;
			}
			//console.log("Adding tile:" + charIndex + "value: " +  itemValue);
			break;
		case "TILELAYOUT":
			if (!lineArray.tileProcessed){
				var arr = lineArray.slice(0).map(ExtractItemValue);
				_CbmTileLayout.setRowArray(importTileRow, arr);
				lineArray.tileProcessed = true;
				importTileRow++;
			}
			break;
		case "AMBIENTCHARS":
			ambientCharImportBuffer[ambientCharIndex] = itemValue;
			ambientCharIndex++;
			break;

		}
	}


	function ExtractToken(line, token) {
		if (line.startsWith(token)) {
			var tokLen = token.length;
			var strCol = line.substr(tokLen).trim();
			return strCol;
		}

		return null;
	}

	function TrySetControl(line, token, control) {
		if (line.startsWith(token)) {
			var tokVal = ExtractToken(line, token);
			$(control).val(tokVal).trigger('change');
		}
	}

	var spareDataItem = null;
	function readRLELine(dataItems) {
		var sourceDataItems = dataItems;
		if (spareDataItem != null) {
			sourceDataItems.unshift(spareDataItem);
			spareDataItem = null;
		}

		var newDataItems = [];
		while (dataItem = sourceDataItems.shift()) {
			extractedDataItem = ExtractItemValue(dataItem);
			if (extractedDataItem >= 128) {
				newDataItems.push(extractedDataItem - 128);
			} else {
				// If an RLE encoded value not truncated....
				if (dataItems.length >= 1) {
					var count = extractedDataItem;
					var itemVal = dataItems.shift();
					var convertedValue = ExtractItemValue(itemVal);
					for (var i = 0; i < count; i++)
						newDataItems.push(itemVal);
				} else {
					// .. else put it back for next time....
					spareDataItem = dataItem;
				}
			}
		}

		return newDataItems;

	}

	function ProcessLine(line) {
		//console.log("Importing line:" + line);

		if (line.startsWith(";rle-encoded=true")) {
			$('#chkRLE').prop('checked', true).triggerHandler('change');
			_rle = true;
			return;
		}

		if (line.startsWith(";char data")) {
			mode = "CHAR";
			//console.log("Mode set to:" + mode);
			return;
		}

		if (line.startsWith(";tiles")) {
			mode = "TILES";
			_CbmTileSet.init();
			//console.log("Mode set to:" + mode);
			return;
		}

		if (line.startsWith(";tile Layout")) {
			mode = "TILELAYOUT";
			//console.log("Mode set to:" + mode);
			return;
		}

		if (line.startsWith(";ambientCharData")) {
			mode = "AMBIENTCHARS";
			//console.log("Mode set to:" + mode);
			return;
		}



		if (line.startsWith(";screen data")) {
			mode = "SCREEN";
			inputScreen = 0;
			screenIndex = 0;
			if (line.startsWith(";screen data1"))
				inputScreen = 1;
			if (line.startsWith(";screen data2"))
				inputScreen = 2;
			if (line.startsWith(";screen data3"))
				inputScreen = 3;
			if (line.startsWith(";screen data4"))
				inputScreen = 4;

			//console.log("Mode set to:" + mode);
			return;
		}

		if (line.startsWith(";colour data")) {
			mode = "COLOUR";
			inputScreen = 0;
			colourIndex = 0;
			if (line.startsWith(";colour data1"))
				inputScreen = 1;
			if (line.startsWith(";colour data2"))
				inputScreen = 2;
			if (line.startsWith(";colour data3"))
				inputScreen = 3;
			if (line.startsWith(";colour data4"))
				inputScreen = 4;
			//console.log("Mode set to:" + mode);
			return;
		}

		// Process geometry changes when not in single mode
		if (!singleScreenOnly) {
			TrySetControl(line, ";background-colour=", "#gBackGroundColourSelect");
			TrySetControl(line, ";border-colour=", "#gBorderColourSelect");
			TrySetControl(line, ";aux-colour=", "#gAuxColourSelect");

			TrySetControl(line, ";col-count=", "#screenColumnCount");
			TrySetControl(line, ";row-count=", "#screenRowCount");

			if (line.startsWith(";char-height=")) {
				var tokVal = ExtractToken(line, ";char-height=");
				//alert(tokVal);
				$('#chkDoubleHeight').prop('checked', tokVal != gCharWidth).triggerHandler('change');
			}
		}

		TrySetControl(line, ";tile char-rows=", "#tileCharSizeRows");
		TrySetControl(line, ";tile char-cols=", "#tileCharSizeCols");
		TrySetControl(line, ";tile rows=", "#tileMapRowCount");
		TrySetControl(line, ";tile cols=", "#tileMapColumnCount");

		var bytePos = line.indexOf(".byte");
		if (bytePos >= 0) {
			//console.log("Reading data line:" + line);

			// Split into data items
			var dataItems = line.substring(bytePos + 6).trim().split(",");

			// Handler RLE
			if (_rle && (mode == "SCREEN" || mode == "COLOUR")) {
				dataItems = readRLELine(dataItems);
			}

			for (var i = 0; i < dataItems.length; i++) {
				ProcessLineItem(i, dataItems[i], dataItems);
			}

		}
	}

	// Generate export text
	var importFile = function (text) {

		charImportBuffer = [];
		ambientCharImportBuffer = [];
		charIndex = 0;
		ambientCharIndex = 0;
		screenIndex = 0;
		colourIndex = 0;
		_CbmTileSet.init();

		singleScreenOnly = $("#chkSingleScreen").is(':checked');

		$('#chkRLE').prop('checked', false).triggerHandler('change');
		_rle = false;

		// Split string into lines
		var lines = text.match(/^.*((\r\n|\n|\r)|$)/gm);

		// For each line
		$.each(lines, function (index, value) {
			ProcessLine(value.trim());
		});

		var charWanted = $("#chkLoadChars").is(':checked');
		if (charWanted)
			uDCGen.setFromBuffer(charImportBuffer);

		// Show radix read from file..
		$("#outputRadix").val(inputRadix);

		primeCharacterTableFromScreen();

		for (var  i=0; i < ambientCharIndex; i++){
			var ambientColor = ambientCharImportBuffer[i];
			var cbmCharObj = uDCGen.getCbmChar(i);
			cbmCharObj.IntendedMultiColor = (ambientColor >= 8);
			cbmCharObj.IntendedCharColor = ambientColor & ~8;
		}

		if (_CbmTileSet.TileCount() < 1)
			_CbmTileSet.newTile();

		QueueDelayedUpdate();
	}

	// Generate export text
	var primeCharacterTableFromScreen = function () {

		// Set attributes on chars based on screen usage..
		for (var iScreen = 0; iScreen < 5; iScreen++) {
			var curScreen = CbmScreens[iScreen];

			// Force to CBM screen for single screen mode...
			if (singleScreenOnly)
				curScreen = CbmScreen;

			// Set colors based on what's in the screen
			for (var i = 0; i < CbmScreen.CharCount; i++) {
				var cidx = curScreen.getCharFromIndex(i);
				var charColour = curScreen.getColValFromIndex(i);
				var cbmCharObj = uDCGen.getCbmChar(cidx);
				cbmCharObj.IntendedMultiColor = (charColour >= 8);
				cbmCharObj.IntendedCharColor = charColour & ~8;
			}
		}

	}

	// Private methods
	var makeTextFile = function (text) {
		var data = new Blob([text], {
				type: 'text/plain'
			});

		// If we are replacing a previously generated file we need to
		// manually revoke the object URL to avoid memory leaks.
		if (textFile !== null) {
			window.URL.revokeObjectURL(textFile);
		}

		textFile = window.URL.createObjectURL(data);
		return textFile;
	};

	function toRadix(num, radix, length) {
		return ("000000000000000" + (Number((num >>> 0)).toString(radix))).slice(-length).toUpperCase()
	}

	function formatOutputValue(value, index) {

		var fmt = parseInt($(outputRadix).val());
		var retString = "";

		switch (fmt) {
		case 0:
			retString = value.toString();
			break
		case 1:
			retString = "0x" + toRadix(value, 16, 2);
			break
		case 2:
			retString = "$" + toRadix(value, 16, 2);
			break
		case 3:
			retString = "%" + toRadix(value, 2, 8);
			break
		}

		return retString;

	}

	var outputAppendArray = function (sourceArray, rleEncode) {

		var outputArray;

		if (rleEncode) {
			outputArray = RLEEncode8(sourceArray);
		} else {
			outputArray = sourceArray;
		}

		for (var i = 0; i < outputArray.length; i++) {
			groupedAppend8(formatOutputValue(outputArray[i]));
		}
	};

	function outputScreen(index, rleWanted) {

		var screen = CbmScreens[index]
			if (!screen.inUse())
				return;

		appendToExport(";screen data" + index);
		appendToExport("ScreenData" + index + ":");
		appendToExport("");

		// Output the screen array
		var outputArray = screen.getCharBuffer();
		outputAppendArray(outputArray, rleWanted);
		groupedAppend8flush();

		groupedAppend8Array = [];
		appendToExport("");
		appendToExport(";colour data" + index);
		appendToExport("ColourDataScreen" + index + ":");
		appendToExport("");

		// Output the screen array
		var outputArray = screen.getColourBuffer();
		outputAppendArray(outputArray, rleWanted);
		groupedAppend8flush();
		appendToExport("");
	}

	// Generate export text
	var generateExport = function (text) {

		var charPrefix = $(charLabelPrefx).val() + "";

		itemsPerLine = 8;
		_exportBuffer = "";
		appendToExport(";settings");
		appendToExport(";background-colour=" + gBackColIdx);
		appendToExport(";border-colour=" + gBorderColIdx);
		appendToExport(";aux-colour=" + gAuxColIdx);

		var rleWanted = $("#chkRLE").is(':checked');
		if (rleWanted) {
			appendToExport(";rle-encoded=" + rleWanted);
		}

		var rows = $("#screenRowCount").val();
		var cols = $("#screenColumnCount").val();
		var normalSettings = true;

		normalSettings &= (gCharHeight == gCharWidth);
		normalSettings &= (rows == 23);
		normalSettings &= (cols == 22);

		if (!normalSettings)
		{
			appendToExport(";char-height=" + gCharHeight);
			appendToExport(";row-count=" + rows);
			appendToExport(";col-count=" + cols);
		}

		appendToExport("");

		appendToExport(";char data");
		appendToExport("CharData:");
		var count = 1 + uDCGen.getHighestModified();
		for (var i = 0; i < count; i++) {
			var charObj = uDCGen.getCbmChar(i);
			var outputValues = $.map(charObj.rows(), formatOutputValue);
			var usedPrefix = "";
			if (charPrefix.length > 1) {
				usedPrefix = charPrefix + i + ": ";
			}
			var output = usedPrefix + ".byte  " + outputValues.join();
			appendToExport(output);
		}

		groupedAppend8Array = [];
		appendToExport("");

		// Output all the screens
		for (var i = 0; i < maxScreens; i++) {
			outputScreen(i, rleWanted);
		}

		outputAmbientCharProps();

		generateTiles();

		generateCharUsageReport();

		return _exportBuffer;
	}

	var makePairsUnique = function (pixelShift) {

		var mappedChars = [];
		var tileCharsX = _CbmTileSet.tileCharColCount;
		var tileCharsY = _CbmTileSet.tileCharRowCount;
		var lastXColUpperBound = tileCharsX - 1;		

		// ToDo: Perform a deep clone of the tilemap so the operation is not destructive
		var clonedLayout = _CbmTileLayout;//.Clone();
		var clonedUDCGen = uDCGen;//.Clone();

		// Split out the tiles to unique tiles
		for (var tileRow = 0; tileRow < clonedLayout.TileRowCount; tileRow++) 
			for (var tileCol = 0;tileCol < clonedLayout.TileColCount; tileCol++) 		
				makeNewTile(clonedLayout, tileRow, tileCol);

		// Next swap the pairs
		for (var tileRow = 0; tileRow < clonedLayout.TileRowCount; tileRow++) {
			for (var tileCharRow = 0; tileCharRow < tileCharsY; tileCharRow++){
				for (var tileCol = 0; tileCol < (clonedLayout.TileColCount - 1); tileCol++) {		
					var tile = clonedLayout.GetTile(tileRow, tileCol);

					for (var tileCharCol = 0; tileCharCol < lastXColUpperBound; tileCharCol++){
						var leftChar = tile.getCharIndex(tileCharRow, tileCharCol);
						var rightChar = tile.getCharIndex(tileCharRow, tileCharCol + 1);
						var replacement = pairInMap(clonedUDCGen, mappedChars, leftChar, rightChar)
						if (replacement){
							tile.setCharIndex(tileCharRow, tileCharCol, replacement);
						}
					}	

					var tileNext = clonedLayout.GetTile(tileRow, tileCol + 1);
					var leftChar = tile.getCharIndex(tileCharRow, lastXColUpperBound);
					var rightChar = tileNext.getCharIndex(tileCharRow, 0);
					var replacement = pairInMap(clonedUDCGen, mappedChars, leftChar, rightChar)
					if (replacement){
						tile.setCharIndex(tileCharRow, lastXColUpperBound, replacement);
					}
				}
			}
		}

		// Set original and target characters
		var shiftedUDCGen = clonedUDCGen;
		var originalChars = [];
		for (var charIdx = 0; charIdx < 255; charIdx++)
			originalChars[charIdx] = shiftedUDCGen.getCbmChar(charIdx).clone();

		var targetChars = [];
		for (var charIdx = 0; charIdx < 255; charIdx++){
			if (pixelShift)
				targetChars[charIdx] = shiftedUDCGen.getCbmChar(charIdx);
			else
				targetChars[charIdx] = shiftedUDCGen.getCbmChar(charIdx).clone();
		}

		// Now shift the chars
		for (var charIndex in mappedChars) {
			var mapping = mappedChars[charIndex];
			var leftCharIndex = mapping.index;
			var rightCharIndex = mapping.twin;
			var targetleftChar = targetChars[leftCharIndex];
			var orgLeftChar = originalChars[leftCharIndex];
			var orgRightChar = originalChars[rightCharIndex];

			for (var row=0; row < gCharHeight; row++){
				targetleftChar.setBit(row, 0, orgLeftChar.getBit(row, 2));
				targetleftChar.setBit(row, 1, orgLeftChar.getBit(row, 3));
				targetleftChar.setBit(row, 2, orgLeftChar.getBit(row, 4));
				targetleftChar.setBit(row, 3, orgLeftChar.getBit(row, 5));
				targetleftChar.setBit(row, 4, orgLeftChar.getBit(row, 6));
				targetleftChar.setBit(row, 5, orgLeftChar.getBit(row, 7));
				targetleftChar.setBit(row, 6, orgRightChar.getBit(row, 0));
				targetleftChar.setBit(row, 7, orgRightChar.getBit(row, 1));
				}
		}

		// Combine to distinct tiles
		var distinctTiles = [];
		var newTileset  = new _CbmTileSetImpl(tileCharsX, tileCharsY);
		for (var tileRow = 0; tileRow < clonedLayout.TileRowCount; tileRow++) {
			for (var tileCol = 0;tileCol < clonedLayout.TileColCount; tileCol++) {
				var tile = clonedLayout.GetTile(tileRow, tileCol);
				var tileSignature = tile.toArray().join();
				if (!distinctTiles[tileSignature]){ 
					var newTile = newTileset.newTile();
					newTile.setFromArray(tile.toArray());
					distinctTiles[tileSignature] = newTile.index;	
				}
				clonedLayout.SetTile(tileRow, tileCol, distinctTiles[tileSignature]);
			}
		}

		_CbmTileSet = newTileset;
		_CbmScreenArray.prepareRelocate();
		_CbmTileSet.prepareRelocate();

		// Work out any duplcates including shifteds
		var uniqueChars = [];
		for (var idx = 0; idx < 255; idx++) {
			var sourceChar = originalChars[idx];
			var sourceShifted = targetChars[idx];
			var sigSource = sourceChar.cbmCharRows.join();
			var sigTarget = sourceShifted.cbmCharRows.join();
			var sigCombined = sigSource + "-" + sigTarget;
			if (uniqueChars[sigCombined]){
				var other = uniqueChars[sigCombined];
				_CbmScreenArray.relocate(idx, other);
				_CbmTileSet.relocate(idx, other);
				uDCGen.getCbmChar(idx).revert();
			} else {
				uniqueChars[sigCombined] = idx;
			}
		}


		alert ("All done!");
		RefreshModifiedTableImages();
		RedrawScreenAndCanvas();
		SetCurCharFocus(gCurrentCanvasChar);	
		RefreshTableCharCount();
		QueueDelayedUpdate();

	}

	function makeNewTile(clonedLayout, tileRow, tileCol){
		var tile = clonedLayout.GetTile(tileRow, tileCol);
		var newTile = _CbmTileSet.newTile();
		newTile.setFromArray(tile.toArray());
		clonedLayout.SetTile(tileRow, tileCol, newTile.index);
	}
	
	function pairInMap(clonedUDCGen, map, leftChar, rightChar){
		//console.log ("Testing pair:" + leftChar + "," + rightChar);

		// Not yet seen
		if (map[leftChar] === undefined){
			map[leftChar] = {index: leftChar, twin: rightChar, mappings: []};
			//console.log ("First seen:" + leftChar + "," + rightChar);
			return null;
		}

		// seen but same paring
		if (map[leftChar].twin === rightChar){
			//console.log ("Duplicate:" + leftChar + "," + rightChar);
			return null;
		}	

		// Left char pointing to a differnt twin....
		
		// Has is already been redirected?

		if (map[leftChar].mappings[rightChar] === undefined){
			var newLeft = clonedUDCGen.createCharClone(leftChar);
			map[leftChar].mappings[rightChar] = newLeft;
			map[newLeft] = {index: newLeft, twin: rightChar, mappings: []};
			console.log ("Creating map:" + leftChar + "," + rightChar + " ==>  " + newLeft + " , " +  rightChar);
			return map[leftChar].mappings[rightChar];
		}

		// Return new right char...
		var preMapped =  map[leftChar].mappings[rightChar];
		//console.log ("reusing  map:" + leftChar + "," + rightChar + " ==>  " + preMapped + " , " +  rightChar);
		return preMapped;
	}


	// Generate export text
	var generateTiles = function (text) {

		appendToExport(";tile data ");
		appendToExport("");

		appendToExport(";tile char-rows=" + _CbmTileSet.tileCharRowCount);
		appendToExport(";tile char-cols=" + _CbmTileSet.tileCharRowCount);
		appendToExport(";tile rows=" + _CbmTileLayout.TileRowCount);
		appendToExport(";tile cols=" + _CbmTileLayout.TileColCount);

		appendToExport(";tiles");
		appendToExport("TileData:");
		// Generate a dictionary of unique char pairs...
		for (var t = 0; t < _CbmTileSet.TileCount(); t++) {
			var tile = _CbmTileSet.getTile(t);
			var bytes = tile.toArray();
			outputAppendArray(bytes);
			groupedAppend8flush();		
		}
		appendToExport("");
		
		itemsPerLine = _CbmTileLayout.TileColCount;
		appendToExport(";tile Layout");
		appendToExport("TileLayoutData:");

		for (var r = 0; r < _CbmTileLayout.TileRowCount; r++) {
			var tiles = _CbmTileLayout.getRowArray(r);
			outputAppendArray(tiles);
			groupedAppend8flush();		
		}

	}

	var outputAmbientCharProps = function () {
		appendToExport("");
		appendToExport(";ambientCharData");
		appendToExport(".if 0");
		var ambientChars = [];
		// Set colors based on what's in the screen
		for (var i = 0; i < uDCGen.getCharCount(); i++) {
			var cbmChar = uDCGen.getCbmChar(i);
			var color = cbmChar.IntendedCharColor;
			if (cbmChar.IntendedMultiColor)
				color|=8;
			ambientChars.push(color);
		}
		outputAppendArray(ambientChars);
		groupedAppend8flush();		
		appendToExport(".endif");
	}

	function fmt3(x) {
		return fmt2(x).replace("00", "--");
	}

	function fmt2(x) {
		return toRadix(x, 10, 2);
	}

	function generateCharUsageReport() {
		appendToExport("");
		appendToExportAsComment("Char Usage Report");
		appendToExportAsComment(" =================                                      00 01 02 03 04");
		var usage = getCharUsage();
		for (var c = 0; c <= 255; c++) {
			var usObj = usage[c];
			var usText = "";
			for (var screen = 0; screen < maxScreens; screen++) {
				usText += fmt3(usObj.ScreenInfo[screen].usage) + ",";
			}

			var outText = "Char Index (" + fmt2(c) + ") TotalScreenCount:" + fmt2(usObj.ScreenCount) + " ScreenUsage [0-4]: " + usText;
			appendToExportAsComment(outText);
		}

	}
	var groupedAppend8Array = [];

	function groupedAppend8flush() {
		appendToExport(".byte  " + groupedAppend8Array.join());
		groupedAppend8Array = [];
	}

	function groupedAppend8(text) {
		if (groupedAppend8Array.length >= itemsPerLine) {
			groupedAppend8flush();
		}
		groupedAppend8Array.push(text);
	}

	function appendToExportAsComment(text) {
		appendToExport("; " + text);
	}

	function appendToExport(text) {
		_exportBuffer += text;
		_exportBuffer += "\r\n";
	}

	// Closure Interface
	return {
		makeTextFile: makeTextFile,
		generateExport: generateExport,
		importFile: importFile,
		primeCharacterTableFromScreen: primeCharacterTableFromScreen,
		makePairsUnique : makePairsUnique
	};

})();
