function blitCharBits(multiColour, grid, fillStyles, canvas, context, cbmChar, x, y) {

	// Pointer into colour registers for nonMC mode
	var row,
	col;
	if (!multiColour) {
		var colRegIdxs = [0, 2];
		for (row = 0; row < gCharHeight; row++) {
			for (col = 0; col < 8; col++) {
				var styleIndex = cbmChar.getBit(row, col);
				context.fillStyle = fillStyles[colRegIdxs[styleIndex]];
				//console.log(fillStyles[0], fillStyles[2]);
				context.fillRect(x + col, y + row, 1, 1);
				if (grid) {
					context.strokeStyle = "grey";
					context.lineWidth = .05;
					context.strokeRect(x + col, y + row, 1, 1);
				} else {
					context.strokeStyle = "white";
					context.lineWidth = 0;
				}
			}
		}
	} else {
		for (row = 0; row < gCharHeight; row++) {
			for (col = 0; col < 8; col += 2) {
				var MSB = cbmChar.getBit(row, col);
				var LSB = cbmChar.getBit(row, col + 1);
				var colorReg = (MSB << 1) + LSB;
				context.fillStyle = fillStyles[colorReg];
				context.fillRect(x + col, y + row, 2, 1);
				if (grid) {
					context.strokeStyle = "grey";
					context.lineWidth = .05;
					context.strokeRect(x + col, y + row, 2, 1);
				} else {
					context.strokeStyle = "white";
					context.lineWidth = 0;
				}
			}
		}
	}
}

function blitCachedCharBits(multiColour, grid, fillStyles, canvas, context, cbmChar, x, y, width, height, screenColorIndex) {

	var bmpCanvas = cbmChar.Canvas(screenColorIndex);

	if (bmpCanvas.width < 1) {
		bmpCanvas.width = width;
		bmpCanvas.height = height;
		var bmCtx = bmpCanvas.getContext("2d");
		bmpCanvas.style.imageRendering = 'pixelated';
		blitCharBits(multiColour, grid, fillStyles, bmpCanvas, bmCtx, cbmChar, 0, 0, bmpCanvas.width, bmpCanvas.height);
	}
	context.imageSmoothingEnabled = false;
	context.drawImage(bmpCanvas, x, y);
}

function prepareCanvas(canvas, context, xSize, ySize, erase) {
	context.setTransform(1, 0, 0, 1, 0, 0);
	var xScale = canvas.width / xSize;
	var yScale = canvas.height / ySize;

	if (erase)
		context.clearRect(0, 0, canvas.width, canvas.height);

	context.scale(xScale, yScale);
}

function RenderTile(canvas, context, tileIndex, showCharGrid, showTileGrid, zoom) {

	var charPerTileX = _CbmTileSet.tileCharColCount;
	var charPerTileY = _CbmTileSet.tileCharRowCount;
	var charPixWidth = gCharWidth;
	var charPixHeight = gCharHeight;

	var tileColsToRender = 1;
	var tileRowsToRender = 1;
	var renderWidthCBMPixels = tileColsToRender * charPerTileX * charPixWidth;
	var renderHeightCBMPixels = tileRowsToRender * charPerTileY * charPixHeight;

	// Selection rect is not inclusive
	prepareCanvas(canvas, context, renderWidthCBMPixels, renderHeightCBMPixels, true);
	context.imageSmoothingEnabled = false;
	canvas.style.imageRendering = 'pixelated';

	DoInnerRenderTile(canvas, context, 0, 0, tileIndex, showCharGrid, showTileGrid);

}

function DoInnerRenderTile(canvas, context, tileRow, tileCol, index, showCharGrid, showTileGrid) {

	var charPerTileX = _CbmTileSet.tileCharColCount;
	var charPerTileY = _CbmTileSet.tileCharRowCount;
	var cbmTile = _CbmTileSet.getTile(index);

	for (var tileCharRow = 0; tileCharRow < charPerTileY; tileCharRow++) {
		for (var tileCharCol = 0; tileCharCol < charPerTileX; tileCharCol++) {

			// Work out if supposed to be multicolour
			var tileCharColour = cbmTile.getCharColor(tileCharRow, tileCharCol);

			// Preserve original value for caching discrimination
			var srctileCharColour = tileCharColour;

			// Determine if multicolor
			var mc = tileCharColour & 8;

			// Mask out multi colour bit for select first 8 characters
			var screenColorIndex = tileCharColour & ~8;

			// Pick correct char colour from  drop down based on screen colour
			var colorStyle = gCharColours[screenColorIndex];

			// Convert tile pos to char pos
			var xCharPos = tileCol * charPerTileX + tileCharCol;
			var yCharPos = tileRow * charPerTileY + tileCharRow;

			// Convert to pixels...
			var xPixelPos = xCharPos * gCharWidth;
			var yPixelPos = yCharPos * gCharHeight;

			// Clear the rect
			context.clearRect(xPixelPos, yPixelPos, gCharWidth, gCharHeight);

			// Obtain the character..
			var cbmCharIndex = cbmTile.getCharIndex(tileCharRow, tileCharCol);
			var cbmChar = uDCGen.getCbmChar(cbmCharIndex);

			var localFillStyles = gmcFillStyles.slice(0);
			var cssCol = gCharColours[cbmChar.IntendedCharColor];
			localFillStyles[2] = cssCol;
			var charGrid = false;

			// Blit the character (color includes multicolr bit)
			blitCachedCharBits(mc, charGrid, localFillStyles, canvas, context, cbmChar, xPixelPos, yPixelPos, gCharWidth, gCharHeight, srctileCharColour);

			if (showCharGrid) {
				context.strokeStyle = "grey";
				context.lineWidth = .032;
				context.strokeRect(xPixelPos, yPixelPos, gCharWidth, gCharHeight);
			}
		}
	}
}

function AddTileImageCellToRow(row, index, showCharGrid, showTileGrid, scaleFactor) {

	// Insert new cells (<td> elements) at the 1st and 2nd position of the "new" <tr> element:
	var cell = row.insertCell(-1);
	var context = tileTempCanvas.getContext("2d");
	RenderTile(tileTempCanvas, context, index, showCharGrid, showTileGrid);
	var imageSource = tileTempCanvas.toDataURL("image/png");
	cell.innerHTML = "";
	var img = document.createElement('img');
	img.height = scaleFactor * tileTempCanvas.height;
	img.src = imageSource;
	img.style.imageRendering = 'pixelated';
	img.tileCellIndex = index;
	img.tableRow = 
	cell.appendChild(img);
	$(img).on('dragstart', DragStartHandler);
	return cell;
}

function RefreshCellImage(img, index, showCharGrid, showTileGrid) {
	var context = tileTempCanvas.getContext("2d");
	RenderTile(tileTempCanvas, context, index, showCharGrid, showTileGrid);
	var imageSource = tileTempCanvas.toDataURL("image/png");
	img.src = imageSource;
	img.style.imageRendering = 'pixelated';
}

function DragStartHandler(e) {

	var json = {
		tileCharIndex: e.target.tileCharIndex,
		tileCellIndex: e.target.tileCellIndex
	};

	var dt = e.originalEvent.dataTransfer;
	dt.setData("text/plain", JSON.stringify(json));
	dt.effectAllowed = "copy";
}
