//Globals
var _tileEditorCanvas = (function () {

	var _charPaletteTableWidth = 8;
	var _tilePaletteTableWidth = 4;
	var _curTileIndex = 0;

	$(document).ready(function () {

		$("#tileEditorCharPaletteTable,#tileEditorTilePalette,#tileEditorTileTable").bind("selectstart", function () {
			return false;
		});
	});

	$("#btnEditTileBits").click(function(e) {
		SetBitmapEditorMode(false);
		_BitMapEditor.onShown(_curTileIndex);
	}); 

	$(document).on(Event_LoadComplete, function(e, hintChar) { 
		init();
	});

	$(document).on(Event_TileGeometryChanged, function(e, hintChar) { 
		init();
	});

	function init(){

		PopulateCharTable($("#tileEditorCharPaletteTable")[0]);
		PopulateTilePaletteTable($("#tileEditorTilePalette")[0]);
		PopulateFocusTileTable($("#tileEditorTileTable")[0]);
	}

	function PopulateFocusTileTable(table) {
		$(table).empty();
		var tile = _CbmTileSet.getTile(_curTileIndex);
		if (!tile)
			return;

		var tmRows = $("#tileCharSizeRows").val();
		var tmCols = $("#tileCharSizeCols").val();
		var cellSize = 120 / tmCols;

		for (var rowIdx = 0; rowIdx < _CbmTileSet.tileCharRowCount; rowIdx++) {
			var row = table.insertRow(-1);
			for (var colIdx = 0; colIdx < _CbmTileSet.tileCharColCount; colIdx++) {
				var charIndex = tile.getCharIndex(rowIdx, colIdx);
				var tableCell = AddCharImageCellToRow(row, charIndex, cellSize);
				$(tableCell).on('dragover', focusCellDragOverHandler);
				$(tableCell).on('drop', focusCellDropHandler);
				// Attach the charIndex to the img
			}
		}
	}

	function PopulateTilePaletteTable(table) {
		$(table).empty();
		var idx = 0;
		var tileCount = _CbmTileSet.TileCount();
		var row;
		while (idx < tileCount) {
			if (idx % _tilePaletteTableWidth === 0)
				row = table.insertRow(-1);
			var cell = AddTileImageCellToRow(row, idx, false, false, 2);
			idx++;
		}
	}

	function PopulateCharTable(table) {
		var startIndex = 0;
		var charCount = uDCGen.getCharCount();
		while (startIndex < charCount) {
			AddRowToCharTable(table, startIndex);
			startIndex += _charPaletteTableWidth;
		}
	}

	function AddRowToCharTable(table, startIndex) {
		// Create an empty <tr> element and add it to the 1st position of the table:
		var row = table.insertRow(-1);
		for (var i = 0; i < _charPaletteTableWidth; i++) {
			AddCharImageCellToRow(row, startIndex + i);
		}
	}

	function AddCharImageCellToRow(row, index, size) {

		// Only add valid indexes..
		if (index >= uDCGen.getCharCount() - 1)
			return;

		// Insert new cells (<td> elements) at the 1st and 2nd position of the "new" <tr> element:
		var cell = row.insertCell(-1);
		doDrawCharCanvas(tempCanvas, index);
		var imageSource = tempCanvas.toDataURL("image/png");
		cell.innerHTML = "";
		var img = document.createElement('img');
		img.src = imageSource;
		img.draggable = true;
		img.style.imageRendering = 'pixelated';
		img.tileCharIndex = index;

		if (size)
			img.height = size;

		// Set image drag drop data..
		$(img).on('dragstart', DragStartHandler);

		cell.appendChild(img);
		return cell;
	}

	function SetCharImage(row, colIdx, charIndex) {
		var cell = row.cells[colIdx];
		doDrawCharCanvas(tempCanvas, charIndex);
		var imageSource = tempCanvas.toDataURL("image/png");
		var img = cell.firstChild;
		img.src = imageSource;
	}

	function RefreshFocusTileTable(table) {
		var tile = _CbmTileSet.getTile(_curTileIndex);
		for (var rowIdx = 0; rowIdx < _CbmTileSet.tileCharRowCount; rowIdx++) {
			var row = table.rows[rowIdx];
			for (var colIdx = 0; colIdx < _CbmTileSet.tileCharColCount; colIdx++) {
				var charIndex = tile.getCharIndex(rowIdx, colIdx);
				SetCharImage(row, colIdx, charIndex);
			}
		}
	}

	// When ambient properties change, only update the char table....
	$(document).on(Event_CharChanged, function(e, hintChar) { 
		handlerCharChanged(hintChar);
	});

	function focusCellDragOverHandler(e) {
		e.preventDefault();
		console.log("Cell drag");
	}

	function focusCellDropHandler(e) {
		console.log("Cell drop");
		var dt = e.originalEvent.dataTransfer;
		var data = dt.getData("text/plain");
		var obj = JSON.parse(data);

		if (obj) {
			if (obj.tileCharIndex !== undefined) {

				// get cell row columns
				var newCharIndex = obj.tileCharIndex;
				var col = $(this).parent().children().index($(this));
				var row = $(this).parent().parent().children().index($(this).parent());

				// Update the tile in the tilelayout
				var tile = _CbmTileSet.getTile(_curTileIndex);
				tile.setCharIndex(row, col, newCharIndex);

				// highlght the cell
				$('#tileEditorTileTable td').removeClass( "charCellHighlight" );
				$(this).addClass( "charCellHighlight" );

				// Attach the charIndex to the img
				var img = $(this)[0].firstChild;
				img.tileCharIndex = newCharIndex;
				
				// Simulate a click to update other views...
				$(img).trigger('click');

				// Inform other views that the cell has changed...
				$(document).trigger(Event_TileChanged, _curTileIndex);

			} else if (obj.tileCellIndex !== undefined) {
				applyNewFocusCell(obj.tileCellIndex);
			}
			//console.log(JSON.parse(data));
		}
	}

	function applyNewFocusCell(index){
		_curTileIndex = index;
		 PopulateFocusTileTable($("#tileEditorTileTable")[0]);
	}

	function RefreshTiles(hintTileIndex) {
		$("#tileEditorTilePalette img").each(function() {
			var img = $(this)[0];
			if (hintTileIndex === undefined)
				hintTileIndex = img.tileCellIndex;
			if (img.tileCellIndex == hintTileIndex)
				RefreshCellImage(img, hintTileIndex);
		});		
	}

	$(document).on(Event_TileChanged, function(e, hintTile) { 
		RefreshFocusTileTable($("#tileEditorTileTable")[0]);
		RefreshTiles(hintTile);
	});

	$(document).on(Event_CharChanged, function(e, hintChar) { 
		handlerCharChanged(hintChar);
	});

	function handlerCharChanged(hintChar){

		var table = $("#tileEditorCharPaletteTable")[0];
		if (table.rows.length < 1)
			return;

		for (var c=0; c < 255; c++){
			if (hintChar !== undefined && hintChar != c)
				continue;
			var charRow = parseInt(c / _charPaletteTableWidth); 
			var charCol = parseInt(c % _charPaletteTableWidth); 
			var row = table.rows[charRow];
			SetCharImage(row, charCol, c); 
		}

	}

	var lastParentHeight = $('#tileEditorDiv').height();
	var lastParentWidth = $('#tileEditorDiv').width();

	$('#actionAddTile').on('click', function () {
		var tile = _CbmTileSet.newTile();
		PopulateTilePaletteTable($("#tileEditorTilePalette")[0]);
		PopulateFocusTileTable($("#tileEditorTileTable")[0]);	
	});

	$('#actionDeleteTile').on('click', function () {
		alert("Not yet implemented!");
		return;
		var tile = _CbmTileSet.removeByIndex(_curTileIndex);
		PopulateTilePaletteTable($("#tileEditorTilePalette")[0]);
		PopulateFocusTileTable($("#tileEditorTileTable")[0]);	
	});

	$('#tileEditorCharPaletteTable,#tileEditorTileTable').on('click', 'td',  function (e) {
		var img = e.target;
		var charIndex = img.tileCharIndex;
		var row = parseInt(charIndex / _charPaletteTableWidth);
		var col = parseInt(charIndex % _charPaletteTableWidth);
		var $cell = $('#tileEditorCharPaletteTable tbody tr').eq(row).find('td').eq(col);
		$('#tileEditorCharPaletteTable td').removeClass( "charCellHighlight" );
		$cell.addClass( "charCellHighlight" );
		$(document).trigger(Event_CharSelected, charIndex);

		$('#tileEditorTileTable td').removeClass( "charCellHighlight" );
		$(this).addClass( "charCellHighlight" );

	});


   
    /* Create 'click' event handler for rows */
    $("#tileEditorTilePalette").on('click', function (e) {
		applyNewFocusCell(e.target.tileCellIndex);
	});

	function resizeHandler() {

		var parentHeight = $('#tileEditorDiv').height();
		var parentWidth = $('#tileEditorDiv').width();
		if (lastParentHeight != parentHeight) {
			updateLayoutSize();
		}
	}
	setInterval(function () {
		resizeHandler();
	}, 100);
})();
