// Represents a single tile
var _CbmTile = function (idx, rows, cols) {

	this.index = idx;

	// Allocate an array to hold the UDC data
	this.tileChars = [];

	// Allocate an array to hold the UDC data
	this.tileColours = [];

	this._copy = [];

	// Initalise the array
	for (var r = 0; r < rows; r++) {
		this.tileChars.push([]);
		var row = this.tileChars[r];
		for (var c = 0; c < cols; c++) {
			row.push(0);
		}
	}
};

// methods _CbmTile (for Represents a single tile)
_CbmTile.prototype = function () {


	var getIndex = function () {
		return this.index;
	};

	var clone = function () {
		var cloned  = new _CbmTile(this.getIndex(), 0,0);
		cloned.setFromArray(this.toArray());
		return cloned;
	};

	var getCharIndex = function (row, col) {
		checkRange(row, col);
		return this.tileChars[row][col];
	};

	var setCharIndex = function (row, col, val) {
		checkRange(row, col);
		this.tileChars[row][col] = val;
	};

	var getCharColor = function (row, col) {
		checkRange(row, col);
		var color = undefined; // TODO: Get/Set color
		
		// If not defined return ambient color for character...	
		if (color === undefined){
			// In the absence of cell colouring use from char palette
			var cbmCharIndex = this.tileChars[row][col];
			var cbmChar = uDCGen.getCbmChar(cbmCharIndex);
			var color = cbmChar.IntendedCharColor;
			var mc = cbmChar.IntendedMultiColor ? 8 : 0;
			return color|mc;
		}
	};

	var getCharMC = function (row, col) {
		checkRange(row, col);
		return getCharColor(row, col).bind(this) & 8;
	};

	var toArray = function () {
		return combineChunks(this.tileChars);
	};

	var prepareRelocate = function () {
		this._copy = this.clone();
	};

	var relocate = function (sourceChar, targetChar) {
		for (var r = 0; r < _CbmTileSet.tileCharRowCount; r++){
			for (var c = 0; c < _CbmTileSet.tileCharColCount; c++){
				var copyCharIndex = this._copy.getCharIndex(r,c);
				if (copyCharIndex == sourceChar)
					setCharIndex.bind(this)(r,c,targetChar);
			}
		}
	};

	var getCharUsageCount = function (charIdx) {
		var charArray = this.toArray();
		return charArray.filter(function(x){return x==charIdx}).length;
	};

	var setFromArray = function (arr) {
		this.tileChars = splitArray(arr, _CbmTileSet.tileCharColCount);
	};

	// ** private methods **

	// Ensure sensible range
	var checkRange = function (row, col) {
		if ((row < 0) || (row > _CbmTileSet.tileCharRowCount))
			throw "row out of range";
		if ((col < 0) || (col > _CbmTileSet.tileCharColCount))
			throw "col out of range";
	};

	// Closure Interface
	return {
		index: this.index,
		getIndex: getIndex,
		getCharIndex: getCharIndex,
		setCharIndex: setCharIndex,
		getCharColor: getCharColor,
		getCharMC : getCharMC,
		toArray: toArray,
		setFromArray: setFromArray,
		relocate : relocate,
		prepareRelocate : prepareRelocate,
		clone : clone,
		getCharUsageCount : getCharUsageCount
	};
}
();

// CbmScreen class
var _CbmTileSetImpl = function (tileCharRows, tileCharCols) {

	var definedTiles = [];
	this.tileCharRowCount = tileCharRows;
	this.tileCharColCount = tileCharCols;

	var init = function () {
		definedTiles = [];
	};

	var TileCount = function () {
		return definedTiles.length;
	};

	var prepareRelocate = function () {
		for (var i = 0; i < TileCount(); i++)
			getTile(i).prepareRelocate();
	};

	var relocate = function (sourceChar, targetChar) {
		for (var i = 0; i < TileCount(); i++)
			getTile(i).relocate(sourceChar, targetChar);
	};

	var newTile = function () {
		var nextTileIdx = TileCount();
		var newTile = new _CbmTile(nextTileIdx, this.tileCharRowCount, this.tileCharColCount);
		definedTiles[nextTileIdx] = newTile;
		return definedTiles[nextTileIdx];
	};

	var getTile = function (idx) {
		if (idx < definedTiles.length)
			return definedTiles[idx];
		else
			return definedTiles[0]; // zero means space not allocated to a tile....
	};

	var getCharUsageCount = function (charIdx) {
		var charUseage = 0;
		for (var i = 0; i < TileCount(); i++)
			charUseage += getTile(i).getCharUsageCount(charIdx);

		return charUseage;
	};

	// Closure Interface
	return {
		tileCharRowCount: this.tileCharRowCount,
		tileCharColCount: this.tileCharColCount,
		TileCount: TileCount,
		newTile: newTile,
		getTile: getTile,
		init: init,
		relocate : relocate,
		prepareRelocate : prepareRelocate,
		getCharUsageCount : getCharUsageCount
	};
};

var _CbmTileSet;
function InitTileSet(rows, cols) {
	_CbmTileSet = new _CbmTileSetImpl(rows, cols);
	_CbmTileSet.newTile();
}

// CbmScreen class
var _CbmTileLayoutImpl = function (rowCount, colCount) {

	var definedCells = [];
	this.tileColCount = colCount;
	this.tileRowCount = rowCount;

	// Initalise the array
	for (var r = 0; r < rowCount; r++) {
		definedCells.push([]);
		var row = definedCells[r];
		for (var c = 0; c < colCount; c++) {
			row.push(0);
		}
	}

	var tileCount = function () {
		return this.TileColCount * this.TileRowCount;
	};

	var getTile = function (row, col) {
		var index = definedCells[row][col];
		return _CbmTileSet.getTile(index);
	};

	var getTileIndex = function (row, col) {
		var index = definedCells[row][col];
		return index;
	};

	var setTile = function (row, col, index) {
		definedCells[row][col] = index;
	};

	var getRowArray = function (row) {
		return definedCells[row].slice(0);
	};

	var setRowArray = function (row, arr) {
		definedCells[row] = arr;
	};

	// Closure Interface
	return {
		SetTile: setTile,
		GetTile: getTile,
		getTileIndex: getTileIndex,
		TileColCount: this.tileColCount,
		TileRowCount: this.tileRowCount,
		TileCount: tileCount,
		getRowArray: getRowArray,
		setRowArray: setRowArray
	};
};

var _CbmTileLayout;
function InitTileMap(rows, cols) {
	_CbmTileLayout = new _CbmTileLayoutImpl(rows, cols);
}

var changedHandler = function (e, hintChar) {
	for (var t = 0; t < _CbmTileSet.TileCount(); t++) {
		var tile = _CbmTileSet.getTile(t);
		if (tile.toArray().indexOf(hintChar) != -1 || hintChar === undefined) {
			// Inform other views that the cell has changed...
			$(document).trigger(Event_TileChanged, t);
		}
	}
};

$(document).on(Event_CharChanged, changedHandler);
