var _tileMapEditor = (function () {

	var showCharGrid = false;
	var showTileGrid = false;
	var scaleValue = 1.0;
	var currentFocusTile = 0;
	var copyBuffer = [];

	function PopulateTileMapTable(table) {
		$(table).empty();
		var idx = 0;
		var rowCount = _CbmTileLayout.TileRowCount;
		for (idx; idx < rowCount; idx++) {
			AddRowToTable(table, idx);
		}
	}


	function AddRowToTable(table, rowIndex, imgSize) {
		var row = table.insertRow(-1);
		var colCount = _CbmTileLayout.TileColCount;
		for (var colIdx = 0; colIdx < colCount; colIdx++) {
			var tile = _CbmTileLayout.GetTile(rowIndex, colIdx);
			if (tile){
				var tableCell = AddTileImageCellToRow(row, tile.index, showCharGrid, showTileGrid, 10);
				$(tableCell).on('dragover', focusCellDragOverHandler);
				$(tableCell).on('drop', focusCellDropHandler);
			}
		}
	}

	function focusCellDragOverHandler(e) {
		e.preventDefault();
		console.log("Table Cell drag");
	}

	function focusCellDropHandler(e) {
		console.log("Cell drop");
		var dt = e.originalEvent.dataTransfer;
		var data = dt.getData("text/plain");
		var obj = JSON.parse(data);

		if (obj) {
			if (obj.tileCellIndex) {

				var colIndex = $(this).parent().children().index($(this));
				var rowIndex = $(this).parent().parent().children().index($(this).parent());
				var existingTile = _CbmTileLayout.GetTile(rowIndex, colIndex);

				// Update tile at that cell if needed
				if (existingTile.index != obj.tileCellIndex) {
					_CbmTileLayout.SetTile(rowIndex, colIndex, obj.tileCellIndex);
					var cell = $(this)[0];
					var img = cell.firstChild;
					img.tileCellIndex = obj.tileCellIndex;
					RefreshCellImage(img, obj.tileCellIndex);
				}
			}
			//console.log(JSON.parse(data));
		}
	}


	function zoomIn() {
		scaleValue = scaleValue * 1.5;
		//scaleValue = Math.max(scaleValue, 1.0);
		$("#tileMapTable").css('transform', 'scale(' + scaleValue + ')');
		return;
	}

	function zoomOut() {
		scaleValue = scaleValue * 0.75;
		//scaleValue = Math.max(scaleValue,  1.0);
		$("#tileMapTable").css('transform', 'scale(' + scaleValue + ')');
		return;
	}

	function RefreshTiles(hintTileIndex) {

		var table = $("#tileMapTable")[0];
		var rowCount = _CbmTileLayout.TileRowCount;
		var colCount = _CbmTileLayout.TileColCount;

		for (var rowIdx = 0; rowIdx < rowCount; rowIdx++) {
			var row = table.rows[rowIdx];
			if (row){
				for (var colIdx = 0; colIdx < colCount; colIdx++) {
					var cell = row.cells[colIdx];
					var img = cell.firstChild;
					img.tileCellIndex = _CbmTileLayout.GetTile(rowIdx, colIdx).index;
					if (img.tileCellIndex == hintTileIndex)
						RefreshCellImage(img, hintTileIndex);
				}
			}
		}
	}

	$(document).on(Event_TileChanged, function (e, hintTile) {
		currentFocusTile = hintTile;
		RefreshTiles(hintTile);
	});

	$(document).on(Event_LoadComplete, function(e, hintChar) { 
		InitTileMapCanvas();
	});

	function InitTileMapCanvas(){
		PopulateTileMapTable($("#tileMapTable")[0]);
		initRowSelect();
	}

	function fillSelection(tileIndex){
		$('#tileMapTable').find(".tmCellSelected").each(function(index) {
			var img = $(this)[0].firstChild;
			img.tileCellIndex = tileIndex;
			var colIndex = $(this).parent().children().index($(this));
			var rowIndex = $(this).parent().parent().children().index($(this).parent());
			var existingTile = _CbmTileLayout.GetTile(rowIndex, colIndex);
			_CbmTileLayout.SetTile(rowIndex, colIndex, tileIndex);
			RefreshCellImage(img, tileIndex);				
		});	
	}
	$('#tmClearSelection').on('click', function () {
		fillSelection(0);
	});

	$('#tmFillSelection').on('click', function () {
		fillSelection(currentFocusTile);
	});

	$('#tmCopySelection').on('click', function () {

		var selectedCells = $('#tileMapTable').find(".tmCellSelected");
		var firstSelectedCell = selectedCells.first();
		if (!selectedCells && firstSelectedCell)
			return;
		
		var cellData  = [];
		var top = firstSelectedCell.parent().parent().children().index(firstSelectedCell.parent());
		var left = firstSelectedCell.parent().children().index(firstSelectedCell);
			
		selectedCells.each(function(index) {
			var img = $(this)[0].firstChild;
			var tileIndex = img.tileCellIndex;
			var colIndex = $(this).parent().children().index($(this));
			var rowIndex = $(this).parent().parent().children().index($(this).parent());
			cellData.push({col:colIndex - left, row:rowIndex - top, tile:tileIndex});
		});	

		if (cellData.length){
			copyBuffer = cellData;
		}
	});
	
	$('#tmPasteSelection').on('click', function () {

		if (!copyBuffer || copyBuffer.length < 1)
			return;

		var selectedCells = $('#tileMapTable').find(".tmCellSelected");
		var firstSelectedCell = selectedCells.first();
		if (!selectedCells || !firstSelectedCell)
			return;
		
		var top = firstSelectedCell.parent().parent().children().index(firstSelectedCell.parent());
		var left = firstSelectedCell.parent().children().index(firstSelectedCell);
		var table = $('#tileMapTable')[0];


		for (i = 0; i< copyBuffer.length; i++) {

			var item = copyBuffer[i];
			var rowIndex = item.row + top;
			var colIndex = item.col + left;
			var tileIndex = item.tile;

			var row = table.rows[rowIndex];
			if (row){
				var cell = row.cells[colIndex];
				if (cell){
					var img = cell.firstChild;
					if (img.tileIndex != tileIndex){
						img.tileCellIndex = tileIndex;
						_CbmTileLayout.SetTile(rowIndex, colIndex, tileIndex);
						RefreshCellImage(img, tileIndex);
					}
				}
			}
		}
	});

	$('#tmSelectRow').on('click', function () {
	});
	$('#tmSelectColumn').on('click', function () {
	});
	$('#tmSelectAll').on('click', function () {
		$('#tileMapTable td').addClass("tmCellSelected");
	});
	


	// ScreenGrid mode handler
	$('#zoomIn').click(zoomIn.bind(this));
	$('#zoomOut').click(zoomOut.bind(this));

	var originalCanvasHeight = $('#tileMapTable').height();
	var originalCanvasWidth = $('#tileMapTable').width();
	var lastParentHeight = $('#tileMapDiv').height();
	var lastPparentWidth = $('#tileMapDiv').width();

	function resizeHandler() {

		var parentHeight = $('#tileMapDiv').height();
		var parentWidth = $('#tileMapDiv').width();

		var updated = false;
		if (lastParentHeight != parentHeight) {
			$('#tileMapTable').height(parentHeight - 160);
			lastParentHeight = $('#tileMapDiv').height();
			updated = true;
		}

		if (lastPparentWidth != parentWidth) {
			$('#tileMapTable').width(parentWidth - 50);
			lastPparentWidth = $('#tileMapDiv').width();
			updated = true;
		}

		if (updated)
			updateLayoutSize();

	}
	setInterval(function () {
		resizeHandler();
	}, 100);

	$(document).on(Event_TileGeometryChanged, function(e, hintChar) { 
		setTimeout(function () {InitTileMapCanvas();}, 500);
	});

	function initRowSelect() {

		var table = $('#tileMapTable');
		var isMouseDown = false;
		var startRowIndex = null;
		var startCellIndex = null;

		function selectTo(cell) {

			var row = cell.parent();
			var cellIndex = cell.index();
			var rowIndex = row.index();

			var rowStart,
			rowEnd,
			cellStart,
			cellEnd;

			if (rowIndex < startRowIndex) {
				rowStart = rowIndex;
				rowEnd = startRowIndex;
			} else {
				rowStart = startRowIndex;
				rowEnd = rowIndex;
			}

			if (cellIndex < startCellIndex) {
				cellStart = cellIndex;
				cellEnd = startCellIndex;
			} else {
				cellStart = startCellIndex;
				cellEnd = cellIndex;
			}

			for (var i = rowStart; i <= rowEnd; i++) {
				var rowCells = table.find("tr").eq(i).find("td");
				for (var j = cellStart; j <= cellEnd; j++) {
					rowCells.eq(j).addClass("tmCellSelected");
				}
			}
		}

		table.find("td").mousedown(function (e) {

			// Drag mode
			var dragMode = $('input:radio[name=tmRadioCurMode]:checked').val();
			if (dragMode > 0)
				return;

			isMouseDown = true;
			var cell = $(this);

			table.find(".tmCellSelected").removeClass("tmCellSelected"); // deselect everything

			if (e.shiftKey) {
				selectTo(cell);
			} else {
				cell.addClass("tmCellSelected");
				startCellIndex = cell.index();
				startRowIndex = cell.parent().index();
			}

			return false; // prevent text selection
		})
		.mouseover(function () {
			if (!isMouseDown)
				return;
			table.find(".tmCellSelected").removeClass("tmCellSelected");
			selectTo($(this));
		})
		.bind("selectstart", function () {
			return false;
		});
		$(document).mouseup(function () {
			isMouseDown = false;
		});
	}

})();
